
/*

TODO...
* expire items in memory cache when it takes up too much memory
*asdfasdf

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "mod_js_config.h"
#include "jscache.h"
#ifndef NGS
#include "jsxdrapi.h"
#endif




static const char cache_directory[]=JAVASCRIPT_LIB_DIR "/cache";
static PRHashTable *the_script_cache=NULL;
static void free_cached_script(Cached_script *this);


/**********
hash table stuff...
*/

static void * CacheAllocTable(void *pool, size_t size) { return malloc(size); }
static void CacheFreeTable(void *pool, void *item) { free(item); }
static PRHashEntry * CacheAllocEntry(void *pool, const void *key)
	{ return malloc(sizeof(PRHashEntry)); }

static void
CacheFreeEntry(void *pool, PRHashEntry *he, uintN flag)
{
	Cached_script *script;
	script=he->value;
	if (flag == HT_FREE_ENTRY) free(he);
	if(script!=NULL) { free_cached_script(script); }
}

static PRHashAllocOps script_cache_hash_ops = {
	CacheAllocTable, CacheFreeTable,
	CacheAllocEntry, CacheFreeEntry
};

/******************
Cache_item...
*/

static void link_cache_item(Cached_script *this,Cache_item *cache_item)
{
	cache_item->next=NULL;
	if(this->first_item==NULL) {
		cache_item->prev=NULL;
		this->last_item=this->first_item=cache_item;
	}
	else {
		/* add this item to the end */
		cache_item->prev=this->last_item;
		this->last_item->next=cache_item;
		this->last_item=cache_item;
	}
}

Cache_item *add_compiled_script_item(Cached_script *this,JSObject *s)
{
	Cache_item *cache_item;

	if((cache_item=JS_malloc(this->cx,sizeof(*cache_item)))==NULL) {
		return NULL;
	}
	cache_item->flags=CACHE_ITEM_SCRIPT;
	cache_item->data.script=s;
	link_cache_item(this,cache_item);
	return cache_item;
}

Cache_item *add_html_item(Cached_script *this,char *html,unsigned int len)
{
	Cache_item *cache_item;

	if((cache_item=JS_malloc(this->cx,sizeof(*cache_item)))==NULL) {
		return NULL;
	}
	cache_item->flags=CACHE_ITEM_HTML;
	cache_item->data.html=html;
	cache_item->length=len;
	link_cache_item(this,cache_item);
	return cache_item;
}


static void free_cache_item(JSContext *cx,Cache_item *this)
{
	if(this->flags&CACHE_ITEM_SCRIPT) {
#if 0
/* ~~~ destroy script item */
		JS_DestroyScript(cx,this->data.script);
#endif
	}
	else if(this->flags&CACHE_ITEM_HTML) {
		JS_free(cx,this->data.html);
	}
	free(this);
}

/******************
XDR...
*/


static void get_filename_in_cache(char *cached_filename,const char *filename)
{
	sprintf(cached_filename,"%s%s",cache_directory,filename);
}

static JSBool make_cache_directories(JSContext *cx,const char *filename)
{
	char directory[FILENAME_MAX];
	const char *p;
	char *to;
	struct stat dir_stat;

	strcpy(directory,cache_directory);
	to=strchr(directory,0);
	for(p=filename; *p; p++) {
		if(*p=='/') {
			*to=0;
			if(stat(directory,&dir_stat)!=0) {
				/* doesn't exist so make it */
				if(mkdir(directory,0770)!=0) {
					JS_ReportError(cx,"Cannot make dir: %s\n");
				}
			}
		}
		*to++=*p;
	}
	return JS_TRUE;
}


/* load/save cached script from xdr
read: only if the cached file is older or old as this one
*/
static JSBool xdr_cached_script(
	JSContext *cx, const char *filename,Cached_script **return_script,
	int mode,int last_modified)
{
	Cached_script *cached_script;
	FILE *cache;
	Cache_item *item_upto;
	struct stat cached_stat;
	char cached_filename[FILENAME_MAX];

	get_filename_in_cache(cached_filename,filename);

	/* go through all the scripts */

	if(mode==JSXDR_DECODE) {
		/* reading: check if it's new enough */
		if(stat(cached_filename,&cached_stat)!=0) { return JS_FALSE; }
		if((last_modified>0 && cached_stat.st_mtime<last_modified)) {
			return JS_FALSE;
		}

		/* make new script to decode into */
		if((cached_script=JS_malloc(cx,sizeof(*cached_script)))==NULL) {
			return JS_FALSE;
		}
		memset(cached_script,0,sizeof(*cached_script));
		cached_script->last_modified=cached_stat.st_mtime;
		cached_script->filename=JS_strdup(cx,filename);
		cached_script->cx=cx;
		cached_script->flags=0;
	}
	else if(mode==JSXDR_ENCODE) {
		/* writing: no need to write into a cached file newer than the script file */
		if(stat(cached_filename,&cached_stat)==0) {
			if((last_modified>0 && cached_stat.st_mtime>=last_modified)) {
				return JS_FALSE;
			}
fprintf(stderr,"Cache is being updated cause file is newer\n");
		}


		/* make the needed directories to write the script */
		cached_script=*return_script;
		make_cache_directories(cached_script->cx,filename);
		item_upto=cached_script->first_item;
	}
	else {
		JS_ReportError(cx,"Unkown mode: %i",mode);
	}

	if((cache=fopen(cached_filename,(mode==JSXDR_ENCODE?"w":"r")))==NULL) {
		JS_ReportError(cx,"Problem opening cache file(%s): %s\n",
		 sys_errlist[errno],cached_filename);
		return JS_FALSE;
	}


	while(1) {
		uint32 flags;

		if(mode==JSXDR_DECODE) {
			if(feof(cache)) { break; }
		}
		else { flags=item_upto->flags; }
		if(JSXDR_ENCODE) { fwrite(&flags,1,sizeof(flags),cache); }
		else { fread(&flags,1,sizeof(flags),cache); }

		if(flags&CACHE_ITEM_SCRIPT) {
			unsigned int len;
			jschar *str;

			if(mode==JSXDR_ENCODE) {
				jsval frozen_jsval;
				JSString *frozen;

fprintf(stderr,"enc111: %x %x\n",cx,item_upto->data.script);
				if(JS_CallFunctionName(cx,item_upto->data.script,"freeze",0,NULL,&frozen_jsval)!=JS_TRUE) { goto bad; }
fprintf(stderr,"frozen\n");
				frozen=JSVAL_TO_STRING(frozen_jsval);
				str=JS_GetStringChars(frozen);
				len=JS_GetStringLength(frozen);
				fwrite(&len,1,sizeof(len),cache);
				fwrite(str,1,len*sizeof(*str),cache);
			}
			else if(mode==JSXDR_DECODE) {
				jsval args[1];
				JSString temp_str;
				jsval rval;
				JSObject *script;

				fread(&len,1,sizeof(len),cache);
				if((str=JS_malloc(cx,len))==NULL) {
					return JS_FALSE;
				}
				fread(str,1,len*sizeof(*str),cache);
				temp_str.chars=str;
				temp_str.length=len;
				args[0]=STRING_TO_JSVAL(&temp_str);
				script=JS_NewScriptObject(cx,NULL);
				if(JS_CallFunctionName(cx,script,"thaw",1,args,&rval)==JS_FALSE) { goto bad; }
				JS_free(cx,str);
				add_compiled_script_item(cached_script,script);
			}
		}
		else if(flags&CACHE_ITEM_HTML) {
			char *str;
			unsigned int len;

			/* add a html item */
			if(mode==JSXDR_ENCODE) {
				len=item_upto->length;
				str=item_upto->data.html;
			}
			else if(mode==JSXDR_DECODE) { 
				fread(&len,1,sizeof(len),cache);
				if((str=JS_malloc(cx,len))==NULL) {
					return JS_FALSE;
				}
				if(fread(str,1,len,cache)!=len) { goto bad; }
				add_html_item(cached_script,str,len);
			}
		}
		else {
			JS_ReportError(cx,"Problem with cache unknown flag: %x",flags);
			goto bad;
		}
		if(mode==JSXDR_ENCODE) {
			item_upto=item_upto->next;
		}
	}
	fclose(cache);
	*return_script=cached_script;
	return JS_TRUE;

bad:;
	fclose(cache);
	if(mode==JSXDR_DECODE) {
		JS_free(cx,cached_script);
	}
	return JS_FALSE;
}


/******************
Cached_script...
*/



static void free_cached_script(Cached_script *this)
{
	Cache_item *i,*next;
	for(i=this->first_item; i!=NULL; ) {
		next=i->next;
		free_cache_item(this->cx,i);
		i=next;
	}
	JS_free(this->cx,this->filename);
	free(this);
}



static Cached_script *lookup_cached_script(const char *filename)
{
	Cached_script *cached_script;
	if((cached_script=PR_HashTableLookup(the_script_cache,filename))!=NULL){
		return cached_script;
	}
	return NULL;
}


/* returns the cached script or a new script
is_new_cached_script to see.
use add_xxxx_item to add items into the cache
*/
Cached_script *get_cached_script(JSRuntime *javascript_runtime,JSContext *cx,const char *filename)
{
	/* look up cached script */
	Cached_script *cached_script;
	struct stat file_stat;

	if(stat(filename,&file_stat)!=0) {	
		fprintf(stderr,"Cannot stat file(%s): %s\n",
			sys_errlist[errno],filename);
		return NULL;
	}
	if((cached_script=lookup_cached_script(filename))!=NULL) {
		if(cached_script->last_modified==file_stat.st_mtime) {
			/* time on file is same as time in cached script,
				script is ok.
			*/

			cached_script->times_runned++;
			return cached_script;
		}
		/* time on script file has changed, delete this script  */
		PR_HashTableRemove(the_script_cache,filename);
	}


	if(cx==NULL) {
		/* setup a javascript context to load script into */
		if((cx=setup_context(javascript_runtime,NULL))==NULL) {
			return NULL;
		}
	}

	/* try to find script in xdr */
	if(xdr_cached_script(cx,filename,&cached_script,JSXDR_DECODE,file_stat.st_mtime)==JS_TRUE){
		cached_script->flags&=-1^CACHED_NEW_SCRIPT;
		return cached_script;
	}

	/* not found or too old: store new script into cache */
	if((cached_script=JS_malloc(cx,sizeof(*cached_script)))==NULL) {
		return NULL;
	}
	memset(cached_script,0,sizeof(cached_script));
	cached_script->last_modified=file_stat.st_mtime;
	cached_script->filename=JS_strdup(cx,filename);
	cached_script->cx=cx;
	cached_script->flags=CACHED_NEW_SCRIPT;

	return cached_script;
}

/* this script is done put it into the cache
*/
void put_script_into_cache(Cached_script *this)
{
	this->flags&=-1^CACHED_NEW_SCRIPT;

	/* put this script into the cache's hash table */
	PR_HashTableAdd(the_script_cache,this->filename,this);

	/* put this script into the file cache */
#if 0
/*~~~ asserts on freezing of script */
	xdr_cached_script(this->cx,this->filename,&this,JSXDR_ENCODE,this->last_modified);
#endif
}


/* get or make the cached script from filename, where the file is a plain javascript */
Cached_script *get_or_make_cached_script(JSContext *cx,const char *filename)
{
	Cached_script *cached_script;
	JSScript *compiled;
	JSObject *script_obj;

	if((cached_script=get_cached_script(NULL,cx,filename))==NULL) {
		return NULL;
	}
	if(!is_new_cached_script(cached_script)) {
		return cached_script;
	}

	if((compiled=JS_CompileFile(cx,JS_GetGlobalObject(cx),filename))==NULL)  { return NULL; }
	script_obj=JS_NewScriptObject(cx,compiled);
	add_compiled_script_item(cached_script,script_obj);
	put_script_into_cache(cached_script);

	return cached_script;
}

/* empty global class */
static JSClass global_class = {
	"global", JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   JS_FinalizeStub
};

/* setup empty context */
JSContext *setup_context(JSRuntime *javascript_runtime, JSContext *cx)
{
	JSObject *glob;

	if(cx==NULL) {
		cx = JS_NewContext(javascript_runtime, 8192);
		if (!cx) return NULL;
		glob = JS_NewObject(cx, &global_class, NULL, NULL);
		if (!glob) { goto fail; }
	}

	if (!JS_InitStandardClasses(cx, glob)
	 || !JS_InitSystemClass(cx,glob)
	 || !JS_InitModuleClass(cx,glob)
	)
		goto fail;
	return cx;
fail:;
	JS_DestroyContext(cx);
	return NULL;
}


JSBool init_script_cache()
{
	the_script_cache=PR_NewHashTable(8,PR_HashString,
		PR_CompareStrings,PR_CompareValues,
		&script_cache_hash_ops,NULL);
	if(the_script_cache==NULL) { return JS_FALSE; }
	return JS_TRUE;
}


