#ifndef __JSCACHE_H
#define __JSCACHE_H

#include "jsmodules.h"

#define CACHE_ITEM_SCRIPT 1
#define CACHE_ITEM_HTML 2
typedef struct Cache_item {
	int flags;
	union {
		JSObject *script;
		char *html;
	}data;
	unsigned int length;
	struct Cache_item *prev,*next;
}Cache_item;

#define CACHED_NEW_SCRIPT 1	/* new script, store script into cache */
typedef struct {
	Cache_item *first_item,*last_item;
	long last_modified;	/* last modified date of this script file */
	char *filename;
	JSContext *cx;		/* each script has a context */
	int times_runned;	/* how many times has this script been ran */
	int flags;
}Cached_script;


typedef struct Context_info {
	Cached_script *cached_script;
}Context_info;

/* check whether this is a cached script to be used or one to be added on */
#define is_new_cached_script(this) (((this)->flags&CACHED_NEW_SCRIPT)?1:0)



Cache_item *add_compiled_script_item(Cached_script *this,JSObject *s);
Cache_item *add_html_item(Cached_script *this,char *html,unsigned int len);


/* returns the cached script or a new script
is_new_cached_script to see.
use add_xxxx_item to add items into the cache
*/
Cached_script *get_cached_script(JSRuntime *javascript_runtime,JSContext *cx,const char *filename);
void put_script_into_cache(Cached_script *this);

Cached_script *get_or_make_cached_script(JSContext *cx,const char *filename);

/* must be runned before anything */
JSBool init_script_cache(void);

/* setup empty context, use NULL or give cx for re-inited context */
JSContext *setup_context(JSRuntime *,JSContext *cx);

#endif
