/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.0 (the "NPL"); you may not use this file except in
 * compliance with the NPL.  You may obtain a copy of the NPL at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the NPL is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the NPL
 * for the specific language governing rights and limitations under the
 * NPL.
 *
 * The Initial Developer of this code under the NPL is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation.  All Rights
 * Reserved.
 */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


#include "mod_js_config.h"
#ifndef NGS
#include <jsscan.h>
#include <jsemit.h>
#include <jsparse.h>
#include <jscntxt.h>
#include <prarena.h>
#include <jsscript.h>
#endif
#include "jscache.h"


const char *
mod_js_getenv (const char *name)
{
    return getenv (name);
}

int
mod_js_setenv (const char *name, const char *value, int overwrite)
{
    return setenv (name, value, overwrite);
}

void
mod_js_env_do(int (*comp) (void *, const char *, const char *), void *rec)
{
    const char **envUpto;
    extern const char **environ;
    const char *equal;
    int rv;

    for(envUpto=environ, rv=1; rv && envUpto[0]!=NULL; envUpto++)
        if((equal=strchr(envUpto[0],'='))!=NULL) {
            char *envNameZ = malloc ((unsigned int)(equal-(envUpto[0])) + 1);

            strncpy(envNameZ,envUpto[0],(unsigned int)(equal-(envUpto[0])));
            envNameZ[equal-envUpto[0]]=0;
            equal++;
            rv = (*comp) (rec, envNameZ, equal);
            free (envNameZ);
        }
        else
            rv = (*comp) (rec, envUpto[0], "");
}

const char **
mod_js_environ (void)
{
    extern const char **environ;

    return environ;
}

void execute_from_stream(JSContext *cx,const char *filename,FILE *in)
{
    JSTokenStream *ts;
    JSCodeGenerator cg;
    JSBool ok;
    JSScript *script;
    jsval result;
    JSString *str;
    JSObject *obj;

    obj=JS_GetGlobalObject(cx);
    ts = js_NewFileTokenStream(cx, filename, in);
    if (!ts)
	goto out;
#ifdef JSDEBUGGER
    if (!filename)
	ts->filename = "typein";
#endif
    if (isatty(fileno(ts->file))) {
	ts->flags |= TSF_INTERACTIVE;
    } else {
        /* Support the UNIX #! shell hack; gobble the first line if it starts
         * with '#'.  TODO - this isn't quite compatible with sharp variables,
         * as a legal js program (using sharp variables) might start with '#'.
         * But that would require multi-character lookahead.
         */
        char ch = fgetc(ts->file);
        if (ch == '#') {
            while((ch = fgetc(ts->file)) != EOF) {
                if(ch == '\n' || ch == '\r')
                    break;
            } 
        }
        ungetc(ch, ts->file);
    }

    do {
	js_InitCodeGenerator(cx, &cg, ts->filename, ts->lineno, ts->principals);
	do {
	    if (ts->flags & TSF_INTERACTIVE)
		printf("js> ");
	    ok = js_CompileTokenStream(cx, obj, ts, &cg);
	    if (ts->flags & TSF_ERROR) {
		ts->flags &= ~TSF_ERROR;
		CLEAR_PUSHBACK(ts);
		ok = JS_TRUE;
	    }
	} while (ok && !(ts->flags & TSF_EOF) && CG_OFFSET(&cg) == 0);
	if (ok) {
	    script = js_NewScriptFromCG(cx, &cg, NULL);
	    if (script) {
		if (JS_ExecuteScript(cx, obj, script, &result) &&
		    (ts->flags & TSF_INTERACTIVE) &&
		    result != JSVAL_VOID) {
		    str = JS_ValueToString(cx, result);
		    if (str)
			printf("%s\n", JS_GetStringBytes(str));
		}
		JS_DestroyScript(cx, script);
	    }
	}
	cg.firstLine = ts->lineno;
	js_ResetCodeGenerator(cx, &cg);
	RESET_TOKENBUF(ts);
    } while (!(ts->flags & TSF_EOF));

out:
    if (ts)
	(void) js_CloseTokenStream(cx, ts);
    PR_FreeArenaPool(&cx->codePool);
    PR_FreeArenaPool(&cx->tempPool);
}

static void jsh_ErrorReporter(JSContext *cx, const char *message, JSErrorReport *report)
{
	fprintf(stderr,"javascript err:");
	if(report!=NULL) {
		if(report->lineno) { fprintf(stderr,"%i:",report->lineno); }
		if(report->filename!=NULL)  { fprintf(stderr,"%s:",report->filename); }
	}
	if(message!=NULL) { fprintf(stderr," %s",message); }
	if(report!=NULL) {
		if(report->linebuf!=NULL) { fprintf(stderr, "\nLine: %s",report->linebuf); }
	}
	fputc('\n',stderr);
}

int main(int argc, char *argv[])
{
	int c;
	char *filename;
	JSRuntime *javascript_runtime;
	JSContext *cx;
	JSScript *script;
	JSObject *global_object;
	jsval rval;
	FILE *in;
	char line[1024];

	filename=NULL;
	for(c=1; c<argc; c++) {
		if(argv[c][0]=='-') {
		}
		else if(filename==NULL) { filename=argv[c]; }
	}

	if(filename==NULL) {
		fprintf(stderr,"no script name mentioned\n");
		exit(1);
	}

	javascript_runtime = JS_Init(8L * 1024L * 1024L);
	if(javascript_runtime==NULL) {
		fprintf(stderr,"javascript: Cannot init runtime");
		exit(1);
	}

	init_modules();
	if((cx=setup_context(javascript_runtime,NULL))==NULL) {
		fprintf(stderr,
			"javascript: Cannot init Javascript: %s",filename);
		exit(1);
	}
	JS_SetErrorReporter(cx, jsh_ErrorReporter);

#ifdef INIT_STATIC_MODULES
	if(0
	 INIT_STATIC_MODULES
	) {
		fprintf(stderr, "javascript: Cannot init static modules\n");
		exit(1);
	}
#endif

	global_object=JS_GetGlobalObject(cx);
	if((in=fopen(filename,"r"))==NULL) {
		fprintf(stderr,"%s: Cannot open script file: %s\n",argv[0],filename);
		exit(1);
	}
	fgets(line,sizeof(line),in);
	if(strncmp(line,"#!",2)==0) {
		execute_from_stream(cx,filename,in);
		fclose(in);
	}
	else {
		fclose(in);

		script=JS_CompileFile(cx,global_object,filename);
		if(JS_ExecuteScript(cx,global_object,script,&rval)!=JS_TRUE) {
			fprintf(stderr,"Problem while executing js script: %s\n",filename);
		}
	}

	JS_DestroyContext(cx);
	JS_DestroyRuntime(javascript_runtime);
	JS_ShutDown();

	return 0;
}

