
/*

TODO...
* in global class anything that doesn't reference loads a module


*/

#include <dlfcn.h>
#include <stdlib.h>
#include <string.h>
#include "mod_js_config.h"


#ifdef JAVASCRIPT_APACHE_MODULE
#include "mod_js.h"
#else
/* only get neccessary config info if not compiling the module */
#include "mod_js_config.h"
#endif



/* Module_handle is created for every time the module is loaded */
typedef struct Module_handle {
	enum {
		MODULE_STATIC,
		MODULE_DYNAMIC,
	}type;
	void *dl_handle;
	Javascript_module *module;
}Module_handle;

/* Module_info is created for every script(context) */
typedef struct {
	Module_handle *module_handle;
	JSObject *module_proto;
}Module_info;

static PRHashTable *loaded_modules;

const char *javacsript_lib_dir=JAVASCRIPT_LIB_DIR;

/**********
hash table stuff...
*/

static void *ModuleAllocTable(void *pool, size_t size) { return malloc(size); }
static void ModuleFreeTable(void *pool, void *item) { free(item); }
static PRHashEntry * ModuleAllocEntry(void *pool, const void *key)
	{ return malloc(sizeof(PRHashEntry)); }

static void
ModuleFreeEntry(void *pool, PRHashEntry *he, uintN flag)
{
	Module_handle *module_handle;
	if (flag == HT_FREE_ENTRY) free(he);
	module_handle=he->value;
	if(module_handle!=NULL) {
		if(module_handle->module!=NULL) {
			free(module_handle->module->js_class->name);
		}
		if(module_handle->type==MODULE_DYNAMIC) {
			dlclose(module_handle->dl_handle);
		}
		free(module_handle);
	}
}

static PRHashAllocOps modules_hash_ops = {
	ModuleAllocTable, ModuleFreeTable,
	ModuleAllocEntry, ModuleFreeEntry
};

/**********
module stuff...
*/

static void finalize_module(JSContext *cx,JSObject *obj)
{
	Module_info *module_info;

	module_info=JS_GetPrivate(cx,obj);
	if(module_info!=NULL) {
		JS_free(cx,module_info);
	}
}

static JSClass module_class = {
	"Module",
	JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   finalize_module
};

static Module_handle *new_module_handle(Javascript_module *module,
	const char *module_name)
{
	Module_handle *module_handle;
	if((module_handle=malloc(sizeof(*module_handle)))==NULL) {
		return NULL;
	}
	module->js_class->name=strdup(module_name);
	module_handle->module=module;
	return module_handle;
}

struct Module_handle *get_module_handle(const char *module_name,
	Javascript_module *static_module)
{
	Javascript_module *module;
	Module_handle *module_handle;
	void *dl_handle;

	/* check if module has been loaded via some other script, load if not */
	if((module_handle=PR_HashTableLookup(loaded_modules,module_name))!=NULL) {
		return module_handle;
	}

	if(static_module!=NULL) {
		module=static_module;
	}
	else {
		char module_filename[FILENAME_MAX];
		char module_struct_name[256];

		/* load the dynamic module */
		sprintf(module_filename, JAVASCRIPT_LIB_DIR "/%s.so",module_name);
		if(!(dl_handle=dlopen(module_filename,RTLD_NOW))) {
			fprintf(stderr,"Cannot load module: %s  (%s)\n",module_name,dlerror());
			return NULL;
		}


		/* get the info from the module */
		sprintf(module_struct_name,"%s_javascript_module",module_name);
		if((module=dlsym(dl_handle,module_struct_name))==NULL) {
			fprintf(stderr,"Cannot get sym link to %s in %s, err:%s\n",
				module_struct_name,module_filename,dlerror());
			return NULL;
		}
	}

	/* save the module info */
	module_handle=new_module_handle(module,module_name);
	module_handle->dl_handle=dl_handle;
	module_handle->type=((static_module!=NULL)?MODULE_STATIC:MODULE_DYNAMIC);
	PR_HashTableAdd(loaded_modules,module_name,module_handle);
	return module_handle;
}


/* load a module and init it
*/
JSObject *load_javascript_module(JSContext *cx, JSObject *obj, const char *module_name)
{
	Javascript_module *module;
	Module_info *module_info;
	JSObject *proto;
	JSObject *module_object;
	jsval module_obj_jsval;
	Module_handle *module_handle;
	const char **needed_modules;

	/* check that the module isn't loaded twice */
	JS_GetProperty(cx,obj,module_name,&module_obj_jsval);
	if(module_obj_jsval!=JSVAL_VOID) {
		module_object=JSVAL_TO_OBJECT(module_obj_jsval);
		module_info=JS_GetPrivate(cx,module_object);
		if(module_info->module_handle->module->init_start!=NULL) {
			module_info->module_handle->module->init_start(cx,module_info->module_proto);
		}
		return module_object;
	}

	if((module_handle=get_module_handle(module_name,NULL))==NULL) {
		return NULL;
	}

	module=module_handle->module;

	/* first, set this module's name into the properties so it 
		can't go on endlessly recursively loading the same modules */
	module_obj_jsval=JSVAL_TRUE;
	JS_SetProperty(cx,obj,module->js_class->name,&module_obj_jsval);

	/* load all the necessary modules needed to run this program */
	needed_modules=module->needed_modules;
	if(needed_modules!=NULL) {
		while(*needed_modules!=NULL) {
			load_javascript_module(cx,obj,*needed_modules);
			needed_modules++;
		}
	}

	/* init the module for this context */
	if(module->init_function!=NULL
	 && (proto= ((*module->init_function)(cx,JS_GetGlobalObject(cx))) )==NULL) {
		JS_DeleteProperty(cx,obj,module->js_class->name);
		return NULL;
	}

	/* module inited ok, save this module's info */
	module_object=JS_NewObject(cx,&module_class,NULL,NULL);
	if((module_info=JS_malloc(cx,sizeof(*module_info)))==NULL) {
		return NULL;
	}
	module_info->module_proto=proto;
	module_info->module_handle=module_handle;
	JS_SetPrivate(cx,module_object,module_info);

	module_obj_jsval=OBJECT_TO_JSVAL(module_object);
	JS_SetProperty(cx,obj,module->js_class->name,&module_obj_jsval);
	return module_object;
}

static JSBool
load_module_method(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	char *module_name;
	JSObject *module_object;
	jsval *argv_upto;
	int c;

	argv_upto=argv;
	for(c=0; c<argc; c++) {
		if(JS_ConvertArguments(cx,1,argv_upto,"s",&module_name)!=JS_TRUE) {
			*rval=JSVAL_FALSE;
			break;
		}

		module_object=load_javascript_module(cx,obj,module_name);
		if(module_object==NULL) {
			*rval=JSVAL_FALSE;
			break;
		}
		argv_upto++;
	}

	*rval=JSVAL_TRUE;
	return JS_TRUE;
}


#if 0
static JSBool
Module(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	/* load a module into the global object */
	char *module_name;

	if(argc<1 || !JSVAL_IS_STRING(argv[0])) {
		return JS_FALSE;
	}

	module_name=JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	return load_javascript_module(cx,obj,module_name);
}
#endif



static JSFunctionSpec module_static_methods[] = {
	{"load",		load_module_method,		1},
	{NULL,NULL,0}
};


JSObject *
JS_InitModuleClass(JSContext *cx, JSObject *obj)
{
	JSObject *proto;

	proto = JS_InitClass(cx, obj, NULL, &module_class, NULL, 0,
		NULL, NULL, NULL,module_static_methods);
	if (!proto) { return NULL; }

	return proto;
}

void init_modules()
{
	loaded_modules=PR_NewHashTable(8,PR_HashString,
		PR_CompareStrings,PR_CompareValues,
		&modules_hash_ops,NULL);
}


