/*
     ``The contents of this file are subject to the Mozilla Public License
     Version 1.0 (the "License"); you may not use this file except in
     compliance with the License. You may obtain a copy of the License at
     http://www.mozilla.org/MPL/

     Software distributed under the License is distributed on an "AS IS"
     basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
     License for the specific language governing rights and limitations
     under the License.

     The Original Code is Copyright (C) 1998 The Apache Group.

     The Initial Developer of the Original Code is Hankin Chick (hankin@dunno.com).

     Contributor(s): .''
*/

/*
* embedded javascript


TODO...
* System.no_cache=1
* System.gets()
* cache plain js as well.
* prefix script to be runned on every script for initing standard modules
* compile option to enable backticks
* System.printf
* Shell module
* standalone

*/


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mod_js_config.h"

#ifndef NGS
#include "prtypes.h"
#include "jscntxt.h"
#include "jsfun.h"
#include "jsgc.h"
#include "jslock.h"
#include "jsparse.h"
#include "jsscan.h"
#include "jsscript.h"
#include "jsconfig.h"
#include "jsemit.h"
#endif

#include "mod_js.h"
#include "jssystem.h"
#include "jscache.h"

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"



/***********************
globals...
*/

request_rec *mod_js_req;

static JSRuntime *javascript_runtime;

/***********************
html stuff...
*/


#define SCRIPT_IS_IN_TAG 0x40000000L	/* script is in tag <?js ?> 
						rather than outside <server></server> */
#define SCRIPT_VERSION_BIT_MASK 0x0fffffffL
/* returns: version number, 0 = use default version, -1 = fail  */
static int is_javascript_tag(const char *tag)
{
	const char *p;
	p=tag;
	while(isspace(*p)) p++;
#if USE_ASP_TAGS
	if(p[0]=='%' && (isspace(p[1]) || p[1]==0 || p[1]=='=')) {
		return SCRIPT_IS_IN_TAG;
	}
	else
#endif
#if USE_PHP_TAGS
	if(strncasecmp(p,"?js",3)==0) {
		return SCRIPT_IS_IN_TAG;
	}
	else
#endif
	if(strncasecmp(p,"server",6)==0) {
		p+=6;
		if(*p!=0 && !isspace(*p)) {
			/* not a space after <server */
			return -1;
		}
		while(*p) {
			if(strncasecmp(p,"language",8)==0) {
				/* check language version */
				int version_required;

				p+=8;
				while(isspace(*p)) p++;
				if(*p=='=') p++;
				while(isspace(*p)) p++;
				if(*p=='\"') p++;
				if(strncasecmp(p,"javascript",10)==0) {
					p+=10;
				}
				else if(*p) {
					/* there is a language tag and it's not javascript */
					return -1;
				}

				version_required=(int)(atof(p)*10);
				if(*p==0 || JS_VERSION >= version_required) {
					return version_required;
				}
				else {
					/* javascript version too high */
					return -1;
				}
			}
			p++;
		}

		/* no language tag, use any language */
		if(*p==0) {
			return 0;
		}
	}
	return -1;
}



/* only works from within find_tag ... */
#define CHECK_DEST_STRING_MEM(len) \
	if((dest_string_upto+(len))>dest_string_end) {	\
		char *new_str;	\
		new_str=malloc(dest_string_len*2);	\
		memcpy(new_str,dest_string[0],dest_string_len);	\
		free(dest_string[0]);	\
		dest_string[0]=new_str;	\
		dest_string_upto=dest_string[0]+dest_string_len;	\
		dest_string_len*=2;	\
		dest_string_end=dest_string[0]+dest_string_len;	\
	}

#define ADD_CHAR(ch) \
	if(dest_string==NULL) { ap_rputc(ch,this); } \
	else { \
		CHECK_DEST_STRING_MEM(1)	\
		*dest_string_upto++=ch;	\
	}

#define ADD_STRING(str,len)	\
	if(dest_string==NULL) { ap_rwrite(str,(len),this); }	\
	else {	\
		CHECK_DEST_STRING_MEM(len);	\
		memcpy(dest_string_upto,str,len);	\
		dest_string_upto+=(len);	\
	}

/* find a string in a stream */
static int find_string_from_stream(request_rec *this,FILE *in,const char *string,
	char **dest_string,int *string_len)
{
	const char *match_upto;
	char *dest_string_upto=NULL,*dest_string_end=NULL;
	int dest_string_len=1024;
	int ch;

	if(dest_string!=NULL) {
		if((dest_string[0]=malloc(dest_string_len))==NULL) {
			return -2;
		}
		dest_string_upto=dest_string[0];
		dest_string_end=dest_string[0]+dest_string_len;
	}

	match_upto=string;
	while((ch=fgetc(in))!=EOF) {
		if(*match_upto==ch) {
			match_upto++;
			if(*match_upto==0) {
				if(dest_string!=NULL) { 
					*dest_string_upto=0; 
					*string_len=dest_string_upto-dest_string[0];
				}
				return 0;
			}
		}
		else {
			if(match_upto!=string) {
				ADD_STRING(string,match_upto-string);
				match_upto=string;
			}
			ADD_CHAR(ch);
		}
	}
	return -1;
}


/* go through the html file till a correct tag is found */
static int find_tag(request_rec *this, FILE *in, int (*is_tag_func)(const char *tag),
	char **dest_string,int *string_len,int *line_number, char **save_tag)
{
	int is_within_comment=0;	/* <!-- comments */
	int is_within_tag=0;		/* within a < ... > */
	int is_within_quote=0;		/* within a " */
	int ch;
	const char *comment_upto;
	char *tag_upto;
	int dest_string_len=1024;
	char *dest_string_upto=NULL,*dest_string_end=NULL;
	char tag[1024];
	static const char comment_start[]="!--";
	static const char comment_end[]="-->";
	char special_tag_end_char=0;

	comment_upto=comment_start;
	if(dest_string!=NULL) {
		if((dest_string[0]=malloc(dest_string_len))==NULL) {
			return -2;
		}
		dest_string_upto=dest_string[0];
		dest_string_end=dest_string[0]+dest_string_len;
	}
	tag_upto=tag;
	while((ch=fgetc(in))!=EOF) {
		if(ch=='\n') { line_number[0]++; }
		if(is_within_comment) {
			ADD_CHAR(ch); 
			/* wait for end of comment */
			if(*comment_upto==ch) {
				comment_upto++;
				if(*comment_upto==0) {
					/* end of comment found */
					is_within_comment=0;
					comment_upto=comment_start;
					is_within_tag=0;
					is_within_quote=0;
				}
			}
			continue;
		}

		if(is_within_tag) {
			if(tag_upto<(tag+sizeof(tag)-1)) {
				if(tag_upto==tag &&
				 (1
#if USE_PHP_TAGS
				 || ch=='?'
#endif
#if USE_ASP_TAGS
				 || ch=='%'
#endif
				 )
				) {
					special_tag_end_char=ch;
				}
				else { special_tag_end_char=0; }

				*tag_upto++=ch;
			}

			if(is_within_quote) {
				/* wait for end of quote */
				if(ch=='\"') { is_within_quote=0; }
				continue;
			}

			/* inside a tag and not within a quote */
			if((special_tag_end_char==0)?
			 (ch=='>')
			 :(tag_upto>tag && tag_upto[-1]==special_tag_end_char && ch=='>') ) {
				int r;
				/* end of tag check if it's a javascript tag */
				*tag_upto=0;
				if((r=is_tag_func(tag))>=0) { 
					if(save_tag!=NULL)  { save_tag[0]=strdup(tag); }
					if(dest_string!=NULL) { 
						*dest_string_upto=0; 
						*string_len=dest_string_upto-dest_string[0];
					}
					return r;
				}
				ADD_CHAR('<');
				ADD_STRING(tag,tag_upto-tag);
				is_within_tag=0;
			}
			else if(ch=='\"') { is_within_quote=1; }
			else {
				int tag_no;

				tag_no=tag_upto-tag;
				if(tag_no==(sizeof(comment_start)-1) 
				 && strncmp(tag,comment_start,sizeof(comment_start)-1)==0) {
					ADD_CHAR('<');
					ADD_STRING(tag,tag_upto-tag);
					is_within_comment=1;
					comment_upto=comment_end;
				}
			}
			continue;
		}

		/* not within a tag */
		if(ch=='<') {
			tag_upto=tag;
			is_within_tag=1;
		}
		else {
			ADD_CHAR(ch);
		}
	}

	/* tag not found */
	if(dest_string!=NULL) { 
		*dest_string_upto=0; 
		*string_len=dest_string_upto-dest_string[0];
	}
	return -1;
}
#undef CHECK_DEST_STRING_MEM
#undef ADD_CHAR
#undef ADD_STRING



/***********************/


const char *
mod_js_getenv (const char *name)
{
  return ap_table_get (mod_js_req->subprocess_env, name);
}

int
mod_js_setenv (const char *name, const char *value, int overwrite)
{
  if (overwrite || !ap_table_get (mod_js_req->subprocess_env, name))
    ap_table_set (mod_js_req->subprocess_env, name, value);
  return 0;
}

void
mod_js_env_do(int (*comp) (void *, const char *, const char *), void *rec)
{
  ap_table_do (comp, rec, mod_js_req->subprocess_env, NULL);
}

const char **
mod_js_environ(void)
{
  return ap_create_environment(mod_js_req->pool,mod_js_req->subprocess_env);
}

/***********************
javascript stuff....
*/



static void apache_ErrorReporter(JSContext *cx, const char *message, JSErrorReport *report)
{
	fprintf(stderr,"javascript err:");
	if(report!=NULL) {
		if(report->lineno) { fprintf(stderr,"%i:",report->lineno); }
		if(report->filename!=NULL)  { fprintf(stderr,"%s:",report->filename); }
	}
	if(message!=NULL) { fprintf(stderr," %s",message); }
	if(report!=NULL) {
		if(report->linebuf!=NULL) { fprintf(stderr, "\nLine: %s",report->linebuf); }
	}
	fputc('\n',stderr);
}


static JSScript *Process(request_rec *this,JSContext *cx, FILE *in,
	char *script_text,int script_len,int line_number)
{
	JSObject *obj;
	JSScript *compiled;
	jsval return_val;

	obj=JS_GetGlobalObject(cx);
	if(script_text==NULL) {
		fseek(in,0,SEEK_END);
		script_len=ftell(in);
		if((script_text=JS_malloc(cx,script_len))==NULL) {
			return NULL;
		}
		fseek(in,0,SEEK_SET);
		fread(script_text,1,script_len,in);
	}

	compiled=JS_CompileScript(cx,obj,script_text,script_len,this->filename,line_number);

	JS_ExecuteScript(cx,obj,compiled,&return_val);

	return compiled;
#if 0
	PR_FreeArenaPool(&cx->codePool);
	PR_FreeArenaPool(&cx->tempPool);
#endif
}


/* reset the info for a new running of stuff */
static void reset_mod_javascript_info(JSContext *cx,request_rec *rec,int status)
{
	Mod_javascript_info *info;

	info=JS_GetPrivate(cx,JS_GetGlobalObject(cx));
	info->request=rec;
	info->status=status;
	info->header_upto=info->header;
}


static Mod_javascript_info *setup_javascript(request_rec *this,JSContext *cx,int version,int status)
{
	JSObject *glob;
	Mod_javascript_info *info;


	if(cx==NULL) {
		if((cx=setup_context(javascript_runtime,cx))==NULL) { return NULL; }
	}
	glob=JS_GetGlobalObject(cx);
	if (version != JSVERSION_DEFAULT) JS_SetVersion(cx, version);

	/* remember the request_rec */
	if((info=JS_malloc(cx,sizeof(*info)))==NULL) { return NULL; }
	memset(info,0,sizeof(*info));
	info->request=this;
	info->status=status;
	info->header_upto=info->header;
	info->cx=cx;
	JS_SetPrivate(cx,glob,info);

	JS_SetErrorReporter(cx, apache_ErrorReporter);

	return info;
}


static int setup_apache_stuff(request_rec *this)
{
	char new_dir[FILENAME_MAX];
	char *p;

	if(this->method_number!=M_GET && this->method_number!=M_POST) {
		return DECLINED;
	}
	ap_add_common_vars(this);
	ap_add_cgi_vars(this);

        mod_js_req = this;

	/* change to the script's dir */
	if((p=strrchr(this->filename,'/'))!=NULL) {
		strncpy(new_dir,this->filename,p-this->filename);
		new_dir[p-this->filename]=0;
		chdir(new_dir);
	}

	return OK;
}


/* find <server tag in html and run the javascript */
static int process_javascript_html(request_rec *this)
{
	FILE *html;
	int apache_ret;
	JSContext *cx;
	Cached_script *cached_script;
	Mod_javascript_info *info;
	int line_number=1;

	if((apache_ret=setup_apache_stuff(this))!=OK) {
		return apache_ret;
	}

	if((cached_script=get_cached_script(javascript_runtime,NULL,this->filename))==NULL) {
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	if(!is_new_cached_script(cached_script)) {
		/* scripts already compiled execute from cache */
		Cache_item *item;

		/* use previous context */
		cx=cached_script->cx;
		reset_mod_javascript_info(cx,this,HTTP_HEADERS_SENT);

		for(item=cached_script->first_item; item!=NULL; item=item->next) {
			if(item->flags&CACHE_ITEM_HTML) {
				ap_rwrite(item->data.html,item->length,this);
			}
			else if(item->flags&CACHE_ITEM_SCRIPT) {
				jsval return_val;
				jsval args[1];

				args[0]=OBJECT_TO_JSVAL(JS_GetGlobalObject(cx));
				if(JS_CallFunctionName(cx,item->data.script,"exec",1,args,&return_val)==JS_FALSE) {
					JS_ReportError(cx,"Problem executing script\n");
				}
			}
		}
		return OK;
	}


	if((info=setup_javascript(this,cached_script->cx,JSVERSION_DEFAULT,HTTP_HEADERS_SENT))==NULL) {
		ap_log_error(APLOG_MARK,APLOG_ERR,this->server,
			"javascript-html: Cannot init Javascript: %s",this->filename);
		return HTTP_INTERNAL_SERVER_ERROR;
	}
	cx=info->cx;
	info->cx_info.cached_script=cached_script;

	if((html=ap_pfopen(this->pool,this->filename,"r"))==NULL) {
		ap_log_error(APLOG_MARK,APLOG_ERR,this->server,
			"javascript: Cannot open file: %s",this->filename);
		return HTTP_INTERNAL_SERVER_ERROR;
	}
	this->content_type="text/html";
	ap_send_http_header(this);

	while(!feof(html)) {
		int version;
		int script_len;
		char *html_text;
		int html_text_len;
		char *script_text=NULL;
		char *tag=NULL;

		/* try to find the tag and save the html into cache */
		version=find_tag(this,html,is_javascript_tag,
		 &html_text,&html_text_len,&line_number,&tag);
		ap_rwrite(html_text,html_text_len,this);
		add_html_item(cached_script,html_text,html_text_len);
		if(version<0) {
			break; 
		}

		if(version&SCRIPT_IS_IN_TAG) {
			/* the script is in the tag itself */
			char *tag_end;

			script_text=tag;
			while(!isspace(*script_text) && *script_text) script_text++;
			tag_end=strchr(tag,0);
			if(tag_end>tag) {
				tag_end--;
				while(tag_end>tag && !isspace(*tag_end)) {
					tag_end--;
				}
				*tag_end=0;
			}
			script_len=strlen(script_text);
		}
		else {
			/* the script is inbetween the start and end tag */
			find_string_from_stream(this,html,"</server>",
				&script_text,&script_len);
		}
		JS_SetVersion(cx, version&SCRIPT_VERSION_BIT_MASK);

		/* run the script */
		if(script_len>0) {
			JSScript *compiled;
			JSObject *script_obj;

			compiled=Process(this,cx,  NULL,script_text,script_len,line_number);
			script_obj=JS_NewScriptObject(cx,compiled);
			/* store in compiled script cache */
			add_compiled_script_item(cached_script,script_obj);
		}

		if(tag!=NULL) { free(tag); }
		else { free(script_text); }
	}

	put_script_into_cache(cached_script);
	ap_pfclose(this->pool,html);
	return OK;
}

/* process a file with plain not-embedded in html javascript */
static int process_javascript(request_rec *this)
{
	FILE *html;
	int apache_ret;
	JSContext *cx;
	Mod_javascript_info *info;

	if((apache_ret=setup_apache_stuff(this))!=OK) {
		return apache_ret;
	}

#if 1
	if((info=setup_javascript(this,NULL,JSVERSION_DEFAULT,0))==NULL) {
		ap_log_error(APLOG_MARK,APLOG_ERR,this->server,
			"javascript: Cannot init Javascript: %s",this->filename);
		return HTTP_INTERNAL_SERVER_ERROR;
	}
	cx=info->cx;

	if((html=ap_pfopen(this->pool,this->filename,"r"))==NULL) {
		ap_log_error(APLOG_MARK,APLOG_ERR,this->server,
			"javascript: Cannot open file: %s",this->filename);
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	Process(this,cx, html,NULL,0,1);

	JS_DestroyContext(cx);
#else

/* ~~~ use cached script here */

#endif
	ap_pfclose(this->pool,html);
	return OK;
}


/*****************************/


static void init_javascript(server_rec *r,pool *p)
{
	javascript_runtime = JS_Init(8L * 1024L * 1024L);
	if(javascript_runtime==NULL) {
		ap_log_error(APLOG_MARK,APLOG_ERR,r,
			"javascript: Cannot init runtime");
	}

	init_modules();
	init_script_cache();

#ifdef INIT_STATIC_MODULES
	if(0
	 INIT_STATIC_MODULES
	) {
		ap_log_error(APLOG_MARK,APLOG_ERR,r,
			"javascript: Cannot init static modules");
	}
#endif
}


static void exit_javascript(server_rec *r,pool *p)
{
	JS_DestroyRuntime(javascript_runtime);
	JS_ShutDown();
}


static handler_rec javascript_handlers[] =
{
	{"javascript-html", process_javascript_html},
	{"javascript-script", process_javascript},
	{NULL}
};


module MODULE_VAR_EXPORT
js_module =
{
    STANDARD_MODULE_STUFF,
    NULL,                       /* initializer */
    NULL, 			/* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    NULL,                       /* server config */
    NULL,                       /* merge server config */
    NULL,			/* command table */
    javascript_handlers,       	/* handlers */
    NULL,                       /* filename translation */
    NULL,                       /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    init_javascript,            /* child_init */
    exit_javascript,            /* child_exit */
    NULL                        /* post read-request */
};


