/* desc: Binary data module


TODO
:
finalize
*/


#include <mod_js_config.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/file.h>
#include <errno.h>

/*************/

typedef struct Binary_private {
	int length;
	char binary[0];
}Binary_private;
static JSClass binary_class;

JSAtom *length_atom;
static const char length_text[]="length";

/*************/

/* add some space to the binary data and return the ptr to start storing stuff */
static char *
add_to_binary(JSContext *cx,Binary_private **binary,int len)
{
	char *next;
	next=binary[0]->binary+binary[0]->length;
	*binary=JS_realloc(cx,*binary,sizeof(*binary)+binary[0]->length+len);
	return next;
}


static JSBool
read_file_method(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	JSString *str;
	Binary_private *binary;
	int c;

	binary=JS_GetPrivate(cx,obj);
	for(c=0; c<argc; c++) {
		int fd;
		char *bin;
		int r;
		struct stat file_stat;
		char *filename;

		if((str=JS_ValueToString(cx,argv[c]))==NULL) {
			JS_ReportError(cx,"cat_file: Cannot get string from %c\n",c);
			return JS_FALSE;
		}
		if((filename=JS_GetStringBytes(str))==NULL) {
			JS_ReportError(cx,"cat_file: Cannot get filename from string: %c\n",c);
			return JS_FALSE;
		}

		if((fd=open(filename,O_RDONLY))<0) {
			JS_ReportError(cx,"cat_file: Cannot open file: %s\n",filename);
			continue;
		}
		if(fstat(fd,&file_stat)<0) {
			JS_ReportError(cx,"cat_file: Cannot stat fd: %i (%s)\n",fd,filename);
		}

		bin=add_to_binary(cx,&binary,file_stat.st_size);

		if((r=read(fd,bin,file_stat.st_size))!=file_stat.st_size) {
			JS_ReportError(cx,"Not enough bytes read: %i\n",r);
		}
		close(fd);
	}
	return JS_TRUE;
}


static JSBool
write_file(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval, int open_mode)
{
	Binary_private *binary;
	const char *filename;
	jsint file_mode=0770;
	int fd;

	if((binary=JS_GetPrivate(cx,obj))==NULL) { return JS_FALSE; }

	if(JS_ConvertArguments(cx,argc,argv,"s",&filename)!=JS_TRUE) {
		JS_ReportError(cx,"write_file: Cannot get filename\n");
		return JS_FALSE;
	}
	if(argc>1) { JS_ValueToInt32(cx,argv[1],&file_mode); }

	if((fd=open(filename,open_mode))>=0) {
		flock(fd,LOCK_EX);
		if(write(fd,binary->binary,binary->length)==binary->length) { *rval=JSVAL_TRUE; }
		else {
			JS_ReportError(cx,"Could not write the whole file: %s (%s)\n",filename,sys_errlist[errno]);
			*rval=JSVAL_FALSE;
		}
		flock(fd,LOCK_UN);
	}
	else {
		JS_ReportError(cx,"Could not open the file: %s (%s)\n",filename,sys_errlist[errno]);
		*rval=JSVAL_FALSE;
	}
	close(fd);

	return JS_TRUE;
}


static JSBool
write_file_method(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	return write_file(cx,obj,argc,argv,rval,O_WRONLY|O_CREAT|O_TRUNC);
}

static JSBool
append_file_method(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	return write_file(cx,obj,argc,argv,rval,O_WRONLY|O_CREAT|O_APPEND);
}


static JSBool
get_binary_property(JSContext *cx, JSObject *obj, jsid id, jsval *rval)
{
	Binary_private *binary;

	if(id==(jsid)length_atom) {
		binary=JS_GetPrivate(cx,obj);
		*rval=INT_TO_JSVAL(binary->length);	
		return JS_TRUE;
	}
	return JS_FALSE;
}


static JSBool
substr_method(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	Binary_private *binary;
	Binary_private *new_binary;
	JSObject *new_obj;
	jsuint pos,len;

	if(JS_ConvertArguments(cx,argc,argv,"uu",&pos,&len)!=JS_TRUE) { return JS_FALSE; }

	if((binary=JS_GetPrivate(cx,obj))==NULL) { return JS_FALSE; }
	if((new_obj=JS_NewObject(cx,&binary_class,NULL,NULL))==NULL) {
		return JS_FALSE;
	}

	new_binary=JS_malloc(cx,sizeof(*new_binary)+len);
	JS_SetPrivate(cx,new_obj,new_binary);
	memcpy(new_binary->binary,binary->binary,len);

	*rval=OBJECT_TO_JSVAL(new_obj);
	return JS_TRUE;
}

/* make a binary object 
*/
static JSBool
Binary(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	int c;
	JSString *str;
	Binary_private *binary;

	binary=JS_malloc(cx,sizeof(*binary));
	memset(binary,0,sizeof(*binary));

	for(c=0; c<argc; c++) {
		jschar *chars,*chars_end;
		char *dest;
		int len;

		if(JSVAL_IS_OBJECT(argv[c])) {
			if(JS_GetClass(obj)==&binary_class) {
				/* is a binary object */
				Binary_private *from;
				JSObject *from_obj;
				char *to;

				from_obj=JSVAL_TO_OBJECT(argv[c]);
				if((from=JS_GetPrivate(cx,from_obj))==NULL) {
					JS_ReportError(cx,"Binary(): Not a binary object\n");
					continue;
				}
				to=add_to_binary(cx,&binary,from->length);
				memcpy(to,from->binary,from->length);
				continue;
			}
		}

		/* add from normal string */
		if((str=JS_ValueToString(cx,argv[c]))==NULL) {
			JS_ReportError(cx,"Cannot get string from argument %c\n",c);
			return JS_FALSE;
		}
		if((chars=JS_GetStringChars(str))==NULL) {
			JS_ReportError(cx,"Binary: Cannot get binary data from argument %c\n",c);
			return JS_FALSE;
		}

		len=JS_GetStringLength(str);
		chars_end=chars+len;
		dest=add_to_binary(cx,&binary,len);
		while(chars<chars_end) {
			*dest++=*chars++;
		}
	}
	JS_SetPrivate(cx,obj,binary);
	*rval=OBJECT_TO_JSVAL(obj);
	return JS_TRUE;
}

static JSBool
binary_to_xxx(JSContext *cx, JSObject *obj, JS_Type type, jsval *rval)
{
	Binary_private *binary;

	binary=JS_GetPrivate(cx,obj);
	if(type==JSTYPE_STRING) {
		JSString *str;
		str=JS_NewStringCopyN(cx,binary->binary,binary->length);
		*rval=STRING_TO_JSVAL(str);
	}
	else {
		return JS_FALSE;
	}
	return JS_TRUE;
}


static JSFunctionSpec binary_methods[] = {
	{"read_file",		read_file_method,		0},
	{"write_file",		write_file_method,		1},
	{"append_file",		append_file_method,		1},
	{"substr",		substr_method,		2},
	{NULL,NULL,0}
};

static JSClass binary_class = {
	NULL,
	JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  get_binary_property,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   binary_to_xxx,   JS_FinalizeStub
};




static JSObject *
InitBinaryModule(JSContext *cx, JSObject *obj)
{
	JSObject *proto;

	proto = JS_InitClass(cx, obj, NULL, &binary_class, Binary, 0,
		NULL, binary_methods, NULL, NULL);
	if (!proto) return NULL;

	length_atom=js_Atomize(cx,length_text,sizeof(length_text)-1,0);

	return proto;
}


Javascript_module Binary_javascript_module={
	&binary_class,
	&InitBinaryModule,
	NULL,
};

