/* Desc: Functions to make common Sql strings
TODO...

*/


#include <mod_js_config.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>

#define is_valid_char(x) ((x)!=';' && (x)!=',' && !isspace((x)))

static JSBool
is_valid_chars(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	int c;
	JSString *from_str;
	jschar *from_end,*from;

	*rval=JSVAL_TRUE;
	for(c=0; c<argc; c++) {
		if(!JSVAL_IS_STRING(argv[c])) { continue; }
		from_str=JSVAL_TO_STRING(argv[c]);
		from_end=from_str->chars+from_str->length;
		for(from=from_str->chars; from<from_end; from++) {
			if(!is_valid_char(*from)) {
				*rval=JSVAL_FALSE;
				return JS_TRUE;
			}
		}
	}

	return JS_TRUE;
}

/*
#define is_quoted_char(ch) ((ch)=='\'' || (ch)=='*' || (ch)=='\"')
*/
#define is_quoted_char(ch) ((ch)=='\'' || (ch)=='\"')

static JSBool
quote_sql(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	int c;
	unsigned int dest_len;
	jschar *dest,*dest_upto,*from_end,*from;
	JSString *from_str;

	/* count how many bytes needed for final string */
	dest_len=0;
	for(c=0; c<argc; c++) {
		if(c!=0) { dest_len+=2; }
		dest_len+=2;	/* leave space for quotes */

		from_str=JSVAL_TO_STRING(argv[c]);
		from_end=from_str->chars+from_str->length;
		dest_len+=from_str->length;
		for(from=from_str->chars; from<from_end; from++) {
			if(is_quoted_char(*from)) { dest_len++; }
		}
	}

	/* make the new string */
	dest=JS_malloc(cx,dest_len*sizeof(*dest));
	dest_upto=dest;
	for(c=0; c<argc; c++) {
		if(c!=0) {
			dest_upto[0]='\'';
			dest_upto[1]=',';
			dest_upto+=2;
		}
		*dest_upto++='\'';
		from_str=JSVAL_TO_STRING(argv[c]);
		from_end=from_str->chars+from_str->length;
		for(from=from_str->chars; from<from_end; from++) {
			int ch;
			ch=*from;
			if(is_quoted_char(ch)) { *dest_upto++='\\'; }
			*dest_upto++=ch;
		}
		*dest_upto++='\'';
	}

	*rval=STRING_TO_JSVAL(JS_NewUCString(cx,dest,dest_len));
	return JS_TRUE;
}

static void
number_to_filename(char *dest, unsigned int number)
{
	sprintf(dest,"%02x/%02x/%02x/%02x",(number>>24)&0xff,(number>>16)&0xff,(number>>8)&0xff,number&0xff);
}


static JSBool
number_to_filename_method(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	jsint n;
	char filename[FILENAME_MAX];
	JSString *str;

	if(JS_ValueToInt32(cx,argv[0],&n)!=JS_TRUE) {
		JS_ReportError(cx,"number_to_filename: is not a number\n");
		return JS_FALSE;
	}
	number_to_filename(filename,n);
	if((str=JS_NewStringCopyZ(cx,filename))==NULL) {
		return JS_FALSE;
	}

	*rval=STRING_TO_JSVAL(str);
	return JS_TRUE;
}


static JSFunctionSpec sql_static_methods[] = {
	{"quote",		quote_sql,		0},
	{"is_valid_chars",	is_valid_chars,		1},
	{"number_to_filename",	number_to_filename_method,		1},
	{NULL,NULL,0}
};



static JSClass sql_class = {
	NULL,
	0,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   JS_FinalizeStub
};




static JSObject *
InitSqlModule(JSContext *cx, JSObject *obj)
{
	JSObject *proto;

	proto = JS_InitClass(cx, obj, NULL, &sql_class, NULL, 0,
		NULL, NULL, NULL, sql_static_methods);
	if (!proto) return NULL;

	return proto;
}


Javascript_module Sql_javascript_module={
	&sql_class,
	&InitSqlModule,
	NULL,
};

