/* Desc: Stuff to handle urls



TODO...
* Url instances: parse url properly
	func: make_url, add 2 urls together, process ../,etc.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mod_js_config.h>


static inline int is_char_allowed(unsigned int ch)
{
	if(isalnum(ch) || ch=='$' || ch=='-' || ch=='_'
	 || ch=='.' || ch=='!' || ch=='~' || ch=='*'
	 || ch=='\'' || ch=='(' || ch==')' || ch==','
/* we don't want to escape paths */
	 || ch=='/'
	) {
		return 1;
	}
	return 0;
}

#if 0
	/* use this one if your CPU's opcode pipeline is alot faster than
		your memory/cache
	*/
#define is_url_char(x) is_char_allowed(ch)
static void init_url_chars { return; }


#else

static unsigned char url_chars_allowed[256];
#define is_url_char(x) (url_chars_allowed[(unsigned char)(x)])
static void init_url_chars(void)
{
	int c;
	for(c=0; c<256; c++) {
		url_chars_allowed[c]=is_char_allowed((unsigned char)c);
	}
}
#endif

static int hex_num[16]={
	'0','1','2','3','4','5','6','7','8','9',
	'A','B','C','D','E','F'
};

static JSBool
url_encode(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	JSString *str;
	jschar *encoded_str,*str_upto,*str_end,*encoded_upto;
	unsigned int encoded_len;

	if(!JSVAL_IS_STRING(argv[0])) {
		str=JS_ValueToString(cx,argv[0]);
	}
	else { str = JSVAL_TO_STRING(argv[0]); }

	/* calculate the length of the encoded string */
	encoded_len=str->length;
	str_upto=str->chars;
	str_end=str_upto+str->length;
	for(; str_upto<str_end; str_upto++) {
		if(*str_upto!=' ' && !is_url_char(*str_upto)) { encoded_len+=2; }
	}
	encoded_str=JS_malloc(cx,encoded_len*sizeof(*encoded_str));
	encoded_upto=encoded_str;
	for(str_upto=str->chars; str_upto<str_end;str_upto++) {
		int ch;
		ch=*str_upto;
		if(ch==' ') {
			*encoded_upto++='+';
		}
		else if(!is_url_char(ch)) {
			*encoded_upto++='%';
			*encoded_upto++=hex_num[(ch>>4)&0xf];
			*encoded_upto++=hex_num[ch&0xf];
		}
		else {
			*encoded_upto++=ch;
		}
	}

	*rval=STRING_TO_JSVAL(JS_NewUCString(cx,encoded_str,encoded_len));

	return JS_TRUE;
}

static JSBool
url_decode(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	JSString *str;
	jschar *encoded,*encoded_end;
	jschar *decoded,*decoded_upto;
	unsigned int length;

	if(JSVAL_IS_STRING(argv[0])) { str=JSVAL_TO_STRING(argv[0]); }
	else { str=JS_ValueToString(cx,argv[0]); }
	length=str->length;

	encoded=str->chars;
	encoded_end=encoded+length;
	for(; encoded<encoded_end; encoded++) {
		if(*encoded=='%') { length-=2; encoded+=2; continue; }
	}

	decoded=JS_malloc(cx,length*sizeof(*decoded));
	decoded_upto=decoded;
	for(encoded=str->chars; encoded<encoded_end; encoded++) {
		jschar ch;
		char hex[3];
		ch=*encoded;
		if(ch=='%') {
			int hex_num;

			hex[0]=encoded[1];
			hex[1]=encoded[2];
			hex[2]=0;
			sscanf(hex,"%02x",&hex_num);
			ch=hex_num;
			encoded+=2;
		}
		else if(ch=='+') { ch=' '; }
		*decoded_upto++=ch;
	}

	*rval=STRING_TO_JSVAL(JS_NewUCString(cx,decoded,length));

	return JS_TRUE;
}

static JSFunctionSpec url_static_methods[] = {
	{"decode",		url_decode,		1},
	{"encode",		url_encode,		1},
	{NULL,NULL,0}
};

static JSClass url_class = {
	NULL,
	0,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   JS_FinalizeStub
};




static JSObject *
InitUrlModule(JSContext *cx, JSObject *obj)
{
	JSObject *proto;

	init_url_chars();
	proto = JS_InitClass(cx, obj, NULL, &url_class, NULL, 0,
		NULL, NULL, NULL, url_static_methods);
	if (!proto) return NULL;

	return proto;
}


Javascript_module Url_javascript_module={
	&url_class,
	&InitUrlModule,
	NULL,
};

